@@ -20,8 +20,8 @@ import threading
 import traceback
 import sys
 import os
+import time
 import subprocess
-import pickle
 import errno
 import hashlib
 import llfuse
@@ -172,7 +172,7 @@ def get_backend_cachedir(storage_url, cachedir):
         raise QuietError('No permission to access cache directory (%s)' % cachedir,
                          exitcode=45)
 
-    return os.path.join(cachedir, _escape(storage_url))
+    return os.path.abspath(os.path.join(cachedir, _escape(storage_url)))
 
 
 def sha256_fh(fh):
@@ -611,11 +611,19 @@ def freeze_basic_mapping(d):
     buf = '{ %s }' % ', '.join(els)
     return buf.encode('utf-8')
 
-def load_params(fname):
-    with open(fname, 'rb') as fh:
-        proto = fh.read(2)
-        fh.seek(0)
-        if proto == b'\x80\x02':
-            return pickle.load(fh)
-        else:
-            return thaw_basic_mapping(fh.read())
+def load_params(cachepath):
+    with open(cachepath + '.params' , 'rb') as fh:
+        return thaw_basic_mapping(fh.read())
+
+def save_params(cachepath, param):
+    with open(cachepath + '.params', 'wb') as fh:
+        fh.write(freeze_basic_mapping(param))
+
+        # Fsync to make sure that the updated sequence number is committed to
+        # disk. Otherwise, a crash immediately after mount could result in both
+        # the local and remote metadata appearing to be out of date.
+        fh.flush()
+        os.fsync(fh.fileno())
+
+def time_ns():
+    return int(time.time() * 1e9)
