@@ -9,7 +9,7 @@ This work can be distributed under the terms of the GNU GPLv3.
 from .logging import logging
 from . import deltadump, CTRL_NAME, CTRL_INODE
 from .backends.common import NoSuchObject, CorruptedObjectError
-from .common import get_path, parse_literal
+from .common import get_path, parse_literal, time_ns
 from .database import NoSuchRowError
 from .inode_cache import OutOfInodesError
 from io import BytesIO
@@ -56,7 +56,6 @@ if not hasattr(errno, 'EOPNOTSUPP'):
 else:
     ACL_ERRNO = errno.EOPNOTSUPP
 
-
 class Operations(llfuse.Operations):
     """A full-featured file system for online data storage
 
@@ -121,7 +120,10 @@ class Operations(llfuse.Operations):
         self.forget(list(self.open_inodes.items()))
         self.inodes.destroy()
 
-    def lookup(self, id_p, name):
+    def lookup(self, id_p, name, ctx):
+        return self._lookup(id_p, name, ctx).entry_attributes()
+
+    def _lookup(self, id_p, name, ctx):
         log.debug('started with %d, %r', id_p, name)
 
         if name == CTRL_NAME:
@@ -151,7 +153,7 @@ class Operations(llfuse.Operations):
         self.open_inodes[inode.id] += 1
         return inode
 
-    def getattr(self, id_):
+    def getattr(self, id_, ctx):
         log.debug('started with %d', id_)
         if id_ == CTRL_INODE:
             # Make sure the control file is only writable by the user
@@ -159,49 +161,44 @@ class Operations(llfuse.Operations):
             inode = self.inodes[CTRL_INODE]
             object.__setattr__(inode, 'uid', os.getuid())
             object.__setattr__(inode, 'gid', os.getgid())
-            return inode
+            return inode.entry_attributes()
 
-        return self.inodes[id_]
+        return self.inodes[id_].entry_attributes()
 
-    def readlink(self, id_):
+    def readlink(self, id_, ctx):
         log.debug('started with %d', id_)
-        timestamp = time.time()
+        now_ns = time_ns()
         inode = self.inodes[id_]
-        if inode.atime < inode.ctime or inode.atime < inode.mtime:
-            inode.atime = timestamp
+        if inode.atime_ns < inode.ctime_ns or inode.atime_ns < inode.mtime_ns:
+            inode.atime_ns = now_ns
         try:
             return self.db.get_val("SELECT target FROM symlink_targets WHERE inode=?", (id_,))
         except NoSuchRowError:
             log.warning('Inode does not have symlink target: %d', id_)
             raise FUSEError(errno.EINVAL)
 
-    def opendir(self, id_):
+    def opendir(self, id_, ctx):
         log.debug('started with %d', id_)
         return id_
 
-    def check_args(self, args):
-        '''Check and/or supplement fuse mount options'''
-
-        args.append(b'big_writes')
-        args.append('max_write=131072')
-        args.append('no_remote_lock')
-
     def readdir(self, id_, off):
         log.debug('started with %d, %d', id_, off)
         if off == 0:
             off = -1
 
         inode = self.inodes[id_]
-        if inode.atime < inode.ctime or inode.atime < inode.mtime:
-            inode.atime = time.time()
+        if inode.atime_ns < inode.ctime_ns or inode.atime_ns < inode.mtime_ns:
+            inode.atime_ns = time_ns()
 
+        # NFS treats offsets 1 and 2 special, so we have to exclude
+        # them.
         with self.db.query("SELECT name_id, name, inode FROM contents_v "
                            'WHERE parent_inode=? AND name_id > ? ORDER BY name_id',
-                           (id_, off)) as res:
+                           (id_, off-3)) as res:
             for (next_, name, cid_) in res:
-                yield (name, self.inodes[cid_], next_)
+                yield (name, self.inodes[cid_].entry_attributes(), next_+3)
 
-    def getxattr(self, id_, name):
+    def getxattr(self, id_, name, ctx):
         log.debug('started with %d, %r', id_, name)
         # Handle S3QL commands
         if id_ == CTRL_INODE:
@@ -226,7 +223,7 @@ class Operations(llfuse.Operations):
                 raise llfuse.FUSEError(llfuse.ENOATTR)
             return value
 
-    def listxattr(self, id_):
+    def listxattr(self, id_, ctx):
         log.debug('started with %d', id_)
         names = list()
         with self.db.query('SELECT name FROM ext_attributes_v WHERE inode=?', (id_,)) as res:
@@ -234,13 +231,14 @@ class Operations(llfuse.Operations):
                 names.append(name)
         return names
 
-    def setxattr(self, id_, name, value):
+    def setxattr(self, id_, name, value, ctx):
         log.debug('started with %d, %r, %r', id_, name, value)
 
         # Handle S3QL commands
         if id_ == CTRL_INODE:
             if name == b's3ql_flushcache!':
-                self.cache.clear()
+                self.inodes.flush()
+                self.cache.flush()
 
             elif name == b'copy':
                 try:
@@ -252,6 +250,7 @@ class Operations(llfuse.Operations):
 
             elif name == b'upload-meta':
                 if self.upload_event is not None:
+                    self.inodes.flush()
                     self.upload_event.set()
                 else:
                     raise llfuse.FUSEError(errno.ENOTTY)
@@ -274,8 +273,7 @@ class Operations(llfuse.Operations):
 
             elif name == b'logging':
                 try:
-                    (lvl, modules)= parse_literal(value, (str, str))
-                    lvl = logging._levelNames[lvl.upper()]
+                    (lvl, modules)= parse_literal(value, (int, str))
                 except (ValueError, KeyError):
                     log.warning('Received malformed command via control inode')
                     raise FUSEError.EINVAL()
@@ -287,6 +285,7 @@ class Operations(llfuse.Operations):
                 except ValueError:
                     log.warning('Received malformed command via control inode')
                     raise FUSEError.EINVAL()
+                log.debug('updated cache size to %d bytes', self.cache.cache.max_size)
 
             else:
                 log.warning('Received unknown command via control inode')
@@ -306,9 +305,9 @@ class Operations(llfuse.Operations):
 
             self.db.execute('INSERT OR REPLACE INTO ext_attributes (inode, name_id, value) '
                             'VALUES(?, ?, ?)', (id_, self._add_name(name), value))
-            self.inodes[id_].ctime = time.time()
+            self.inodes[id_].ctime_ns = time_ns()
 
-    def removexattr(self, id_, name):
+    def removexattr(self, id_, name, ctx):
         log.debug('started with %d, %r', id_, name)
 
         if self.failsafe or self.inodes[id_].locked:
@@ -324,7 +323,7 @@ class Operations(llfuse.Operations):
         if changes == 0:
             raise llfuse.FUSEError(llfuse.ENOATTR)
 
-        self.inodes[id_].ctime = time.time()
+        self.inodes[id_].ctime_ns = time_ns()
 
     def lock_tree(self, id0):
         '''Lock directory tree'''
@@ -336,7 +335,7 @@ class Operations(llfuse.Operations):
         queue = [ id0 ]
         self.inodes[id0].locked = True
         processed = 0 # Number of steps since last GIL release
-        stamp = time.time() # Time of last GIL release
+        stamp = time_ns() # Time of last GIL release
         gil_step = 250 # Approx. number of steps between GIL releases
         while True:
             id_p = queue.pop()
@@ -374,7 +373,7 @@ class Operations(llfuse.Operations):
         if self.inodes[id_p0].locked:
             raise FUSEError(errno.EPERM)
 
-        id0 = self.lookup(id_p0, name0).id
+        id0 = self._lookup(id_p0, name0, ctx=None).id
         queue = [ id0 ] # Directories that we still need to delete
         processed = 0 # Number of steps since last GIL release
         stamp = time.time() # Time of last GIL release
@@ -443,8 +442,7 @@ class Operations(llfuse.Operations):
         db = self.db
 
         # First we make sure that all blocks are in the database
-        self.cache.commit()
-        log.debug('committed cache')
+        self.cache.start_flush()
 
         # Copy target attributes
         # These come from setxattr, so they may have been deleted
@@ -454,13 +452,13 @@ class Operations(llfuse.Operations):
             target_inode = self.inodes[target_id]
         except KeyError:
             raise FUSEError(errno.ENOENT)
-        for attr in ('atime', 'ctime', 'mtime', 'mode', 'uid', 'gid'):
+        for attr in ('atime_ns', 'ctime_ns', 'mtime_ns', 'mode', 'uid', 'gid'):
             setattr(target_inode, attr, getattr(src_inode, attr))
 
         # We first replicate into a dummy inode, so that we
         # need to invalidate only once.
-        timestamp = time.time()
-        tmp = make_inode(mtime=timestamp, ctime=timestamp, atime=timestamp,
+        now_ns = time_ns()
+        tmp = make_inode(mtime_ns=now_ns, ctime_ns=now_ns, atime_ns=now_ns,
                          uid=0, gid=0, mode=0, refcount=0)
 
         queue = [ (src_id, tmp.id, 0) ]
@@ -481,8 +479,8 @@ class Operations(llfuse.Operations):
                         try:
                             inode_new = make_inode(refcount=1, mode=inode.mode, size=inode.size,
                                                    uid=inode.uid, gid=inode.gid,
-                                                   mtime=inode.mtime, atime=inode.atime,
-                                                   ctime=inode.ctime, rdev=inode.rdev)
+                                                   mtime_ns=inode.mtime_ns, atime_ns=inode.atime_ns,
+                                                   ctime_ns=inode.ctime_ns, rdev=inode.rdev)
                         except OutOfInodesError:
                             log.warning('Could not find a free inode')
                             raise FUSEError(errno.ENOSPC)
@@ -546,12 +544,12 @@ class Operations(llfuse.Operations):
 
         log.debug('finished')
 
-    def unlink(self, id_p, name):
+    def unlink(self, id_p, name, ctx):
         log.debug('started with %d, %r', id_p, name)
         if self.failsafe:
             raise FUSEError(errno.EPERM)
 
-        inode = self.lookup(id_p, name)
+        inode = self._lookup(id_p, name, ctx)
 
         if stat.S_ISDIR(inode.mode):
             raise llfuse.FUSEError(errno.EISDIR)
@@ -560,12 +558,12 @@ class Operations(llfuse.Operations):
 
         self.forget([(inode.id, 1)])
 
-    def rmdir(self, id_p, name):
+    def rmdir(self, id_p, name, ctx):
         log.debug('started with %d, %r', id_p, name)
         if self.failsafe:
             raise FUSEError(errno.EPERM)
 
-        inode = self.lookup(id_p, name)
+        inode = self._lookup(id_p, name, ctx)
 
         if self.inodes[id_p].locked:
             raise FUSEError(errno.EPERM)
@@ -588,7 +586,7 @@ class Operations(llfuse.Operations):
 
         log.debug('started with %d, %r', id_p, name)
 
-        timestamp = time.time()
+        now_ns = time_ns()
 
         # Check that there are no child entries
         if self.db.has_val("SELECT 1 FROM contents WHERE parent_inode=?", (id_,)):
@@ -605,11 +603,11 @@ class Operations(llfuse.Operations):
 
         inode = self.inodes[id_]
         inode.refcount -= 1
-        inode.ctime = timestamp
+        inode.ctime_ns = now_ns
 
         inode_p = self.inodes[id_p]
-        inode_p.mtime = timestamp
-        inode_p.ctime = timestamp
+        inode_p.mtime_ns = now_ns
+        inode_p.ctime_ns = now_ns
 
         if inode.refcount == 0 and id_ not in self.open_inodes:
             log.debug('removing from cache')
@@ -647,9 +645,9 @@ class Operations(llfuse.Operations):
         self.db.execute('INSERT INTO symlink_targets (inode, target) VALUES(?,?)',
                         (inode.id, target))
         self.open_inodes[inode.id] += 1
-        return inode
+        return inode.entry_attributes()
 
-    def rename(self, id_p_old, name_old, id_p_new, name_new):
+    def rename(self, id_p_old, name_old, id_p_new, name_new, ctx):
         log.debug('started with %d, %r, %d, %r', id_p_old, name_old, id_p_new, name_new)
         if name_new == CTRL_NAME or name_old == CTRL_NAME:
             log.warning('Attempted to rename s3ql control file (%s -> %s)',
@@ -662,10 +660,10 @@ class Operations(llfuse.Operations):
             or self.inodes[id_p_new].locked):
             raise FUSEError(errno.EPERM)
 
-        inode_old = self.lookup(id_p_old, name_old)
+        inode_old = self._lookup(id_p_old, name_old, ctx)
 
         try:
-            inode_new = self.lookup(id_p_new, name_new)
+            inode_new = self._lookup(id_p_new, name_new, ctx)
         except llfuse.FUSEError as exc:
             if exc.errno != errno.ENOENT:
                 raise
@@ -714,7 +712,7 @@ class Operations(llfuse.Operations):
         return name_id
 
     def _rename(self, id_p_old, name_old, id_p_new, name_new):
-        timestamp = time.time()
+        now_ns = time_ns()
 
         name_id_new = self._add_name(name_new)
         name_id_old = self._del_name(name_old)
@@ -725,15 +723,15 @@ class Operations(llfuse.Operations):
 
         inode_p_old = self.inodes[id_p_old]
         inode_p_new = self.inodes[id_p_new]
-        inode_p_old.mtime = timestamp
-        inode_p_new.mtime = timestamp
-        inode_p_old.ctime = timestamp
-        inode_p_new.ctime = timestamp
+        inode_p_old.mtime_ns = now_ns
+        inode_p_new.mtime_ns = now_ns
+        inode_p_old.ctime_ns = now_ns
+        inode_p_new.ctime_ns = now_ns
 
     def _replace(self, id_p_old, name_old, id_p_new, name_new,
                  id_old, id_new):
 
-        timestamp = time.time()
+        now_ns = time_ns()
 
         if self.db.has_val("SELECT 1 FROM contents WHERE parent_inode=?", (id_new,)):
             log.info("Attempted to overwrite entry with children: %s",
@@ -752,15 +750,15 @@ class Operations(llfuse.Operations):
 
         inode_new = self.inodes[id_new]
         inode_new.refcount -= 1
-        inode_new.ctime = timestamp
+        inode_new.ctime_ns = now_ns
 
         inode_p_old = self.inodes[id_p_old]
-        inode_p_old.ctime = timestamp
-        inode_p_old.mtime = timestamp
+        inode_p_old.ctime_ns = now_ns
+        inode_p_old.mtime_ns = now_ns
 
         inode_p_new = self.inodes[id_p_new]
-        inode_p_new.ctime = timestamp
-        inode_p_new.mtime = timestamp
+        inode_p_new.ctime_ns = now_ns
+        inode_p_new.mtime_ns = now_ns
 
         if inode_new.refcount == 0 and id_new not in self.open_inodes:
             self.cache.remove(id_new, 0,
@@ -776,7 +774,7 @@ class Operations(llfuse.Operations):
             del self.inodes[id_new]
 
 
-    def link(self, id_, new_id_p, new_name):
+    def link(self, id_, new_id_p, new_name, ctx):
         log.debug('started with %d, %d, %r', id_, new_id_p, new_name)
 
         if new_name == CTRL_NAME or id_ == CTRL_INODE:
@@ -784,7 +782,7 @@ class Operations(llfuse.Operations):
                       get_path(new_id_p, self.db, new_name))
             raise llfuse.FUSEError(errno.EACCES)
 
-        timestamp = time.time()
+        now_ns = time_ns()
         inode_p = self.inodes[new_id_p]
 
         if inode_p.refcount == 0:
@@ -795,33 +793,49 @@ class Operations(llfuse.Operations):
         if self.failsafe or inode_p.locked:
             raise FUSEError(errno.EPERM)
 
-        inode_p.ctime = timestamp
-        inode_p.mtime = timestamp
+        inode_p.ctime_ns = now_ns
+        inode_p.mtime_ns = now_ns
 
         self.db.execute("INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)",
                         (self._add_name(new_name), id_, new_id_p))
         inode = self.inodes[id_]
         inode.refcount += 1
-        inode.ctime = timestamp
+        inode.ctime_ns = now_ns
 
         self.open_inodes[inode.id] += 1
-        return inode
+        return inode.entry_attributes()
 
-    def setattr(self, id_, attr):
+    def setattr(self, id_, attr, fields, fh, ctx):
         """Handles FUSE setattr() requests"""
-        if log.isEnabledFor(logging.DEBUG):
-            log.debug('started with %d, %s', id_,
-                      ', '.join('%s=%r' % (x, getattr(attr, x))
-                                for x in dir(attr)
-                                if x.startswith('st_') and getattr(attr, x) is not None ))
-
         inode = self.inodes[id_]
-        timestamp = time.time()
+        if fh is not None:
+            assert fh == id_
+        now_ns = time_ns()
 
         if self.failsafe or inode.locked:
             raise FUSEError(errno.EPERM)
 
-        if attr.st_size is not None:
+        if fields.update_mode:
+            inode.mode = attr.st_mode
+
+        if fields.update_uid:
+            inode.uid = attr.st_uid
+
+        if fields.update_gid:
+            inode.gid = attr.st_gid
+
+        if fields.update_atime:
+            inode.atime_ns = attr.st_atime_ns
+
+        if fields.update_mtime:
+            inode.mtime_ns = attr.st_mtime_ns
+
+        inode.ctime_ns = now_ns
+
+        # This needs to go last, because the call to cache.remove and cache.get
+        # will release the global lock and may thus evict the *inode* object
+        # from the cache.
+        if fields.update_size:
             len_ = attr.st_size
 
             # Determine blocks to delete
@@ -852,24 +866,8 @@ class Operations(llfuse.Operations):
                     self.broken_blocks[id_].add(last_block)
                     raise FUSEError(errno.EIO)
 
-        if attr.st_mode is not None:
-            inode.mode = attr.st_mode
 
-        if attr.st_uid is not None:
-            inode.uid = attr.st_uid
-
-        if attr.st_gid is not None:
-            inode.gid = attr.st_gid
-
-        if attr.st_atime is not None:
-            inode.atime = attr.st_atime
-
-        if attr.st_mtime is not None:
-            inode.mtime = attr.st_mtime
-
-        inode.ctime = timestamp
-
-        return inode
+        return inode.entry_attributes()
 
     def mknod(self, id_p, name, mode, rdev, ctx):
         log.debug('started with %d, %r', id_p, name)
@@ -877,7 +875,7 @@ class Operations(llfuse.Operations):
             raise FUSEError(errno.EPERM)
         inode = self._create(id_p, name, mode, ctx, rdev=rdev)
         self.open_inodes[inode.id] += 1
-        return inode
+        return inode.entry_attributes()
 
     def mkdir(self, id_p, name, mode, ctx):
         log.debug('started with %d, %r', id_p, name)
@@ -885,7 +883,7 @@ class Operations(llfuse.Operations):
             raise FUSEError(errno.EPERM)
         inode = self._create(id_p, name, mode, ctx)
         self.open_inodes[inode.id] += 1
-        return inode
+        return inode.entry_attributes()
 
     def extstat(self):
         '''Return extended file system statistics'''
@@ -900,17 +898,18 @@ class Operations(llfuse.Operations):
         inodes = self.db.get_val("SELECT COUNT(id) FROM inodes")
         fs_size = self.db.get_val('SELECT SUM(size) FROM inodes') or 0
         dedup_size = self.db.get_val('SELECT SUM(size) FROM blocks') or 0
-        compr_size = self.db.get_val('SELECT SUM(size) FROM objects') or 0
-        (cache_used, cache_dirty) = self.cache.get_usage()
 
-        return struct.pack('QQQQQQQQQ', entries, blocks, inodes, fs_size, dedup_size,
-                           compr_size, self.db.get_size(), cache_used, cache_dirty)
+        # Objects that are currently being uploaded/compressed have size == -1
+        compr_size = self.db.get_val('SELECT SUM(size) FROM objects '
+                                     'WHERE size > 0') or 0
 
+        return struct.pack('QQQQQQQQQQQQ', entries, blocks, inodes, fs_size, dedup_size,
+                           compr_size, self.db.get_size(), *self.cache.get_usage())
 
-    def statfs(self):
+    def statfs(self, ctx):
         log.debug('started')
 
-        stat_ = llfuse.StatvfsData
+        stat_ = llfuse.StatvfsData()
 
         # Get number of blocks & inodes
         blocks = self.db.get_val("SELECT COUNT(id) FROM objects")
@@ -947,7 +946,7 @@ class Operations(llfuse.Operations):
 
         return stat_
 
-    def open(self, id_, flags):
+    def open(self, id_, flags, ctx):
         log.debug('started with %d', id_)
         if ((flags & os.O_RDWR or flags & os.O_WRONLY)
             and (self.failsafe or self.inodes[id_].locked)):
@@ -979,11 +978,11 @@ class Operations(llfuse.Operations):
         except NoSuchRowError:
             inode = self._create(id_p, name, mode, ctx)
         else:
-            self.open(id_, flags)
+            self.open(id_, flags, ctx)
             inode = self.inodes[id_]
 
         self.open_inodes[inode.id] += 1
-        return (inode.id, inode)
+        return (inode.id, inode.entry_attributes())
 
     def _create(self, id_p, name, mode, ctx, rdev=0, size=0):
         if name == CTRL_NAME:
@@ -991,7 +990,7 @@ class Operations(llfuse.Operations):
                      get_path(id_p, self.db, name))
             raise FUSEError(errno.EACCES)
 
-        timestamp = time.time()
+        now_ns = time_ns()
         inode_p = self.inodes[id_p]
 
         if inode_p.locked:
@@ -1001,12 +1000,18 @@ class Operations(llfuse.Operations):
             log.warning('Attempted to create entry %s with unlinked parent %d',
                      name, id_p)
             raise FUSEError(errno.EINVAL)
-        inode_p.mtime = timestamp
-        inode_p.ctime = timestamp
+        inode_p.mtime_ns = now_ns
+        inode_p.ctime_ns = now_ns
 
+        if inode_p.mode & stat.S_ISGID:
+            gid = inode_p.gid
+            if stat.S_ISDIR(mode):
+                mode |= stat.S_ISGID
+        else:
+            gid = ctx.gid
         try:
-            inode = self.inodes.create_inode(mtime=timestamp, ctime=timestamp, atime=timestamp,
-                                             uid=ctx.uid, gid=ctx.gid, mode=mode, refcount=1,
+            inode = self.inodes.create_inode(mtime_ns=now_ns, ctime_ns=now_ns, atime_ns=now_ns,
+                                             uid=ctx.uid, gid=gid, mode=mode, refcount=1,
                                              rdev=rdev, size=size)
         except OutOfInodesError:
             log.warning('Could not find a free inode')
@@ -1044,8 +1049,8 @@ class Operations(llfuse.Operations):
         # Inode may have expired from cache
         inode = self.inodes[fh]
 
-        if inode.atime < inode.ctime or inode.atime < inode.mtime:
-            inode.atime = time.time()
+        if inode.atime_ns < inode.ctime_ns or inode.atime_ns < inode.mtime_ns:
+            inode.atime_ns = time_ns()
 
         return buf.getvalue()
 
@@ -1072,11 +1077,11 @@ class Operations(llfuse.Operations):
         # so we have to be careful not to undo a size extension made by
         # a concurrent write (because _readwrite() releases the global
         # lock).
-        timestamp = time.time()
+        now_ns = time_ns()
         inode = self.inodes[fh]
         inode.size = max(inode.size, minsize)
-        inode.mtime = timestamp
-        inode.ctime = timestamp
+        inode.mtime_ns = now_ns
+        inode.ctime_ns = now_ns
 
         return total
 
@@ -1148,7 +1153,7 @@ class Operations(llfuse.Operations):
             self.inodes.flush_id(fh)
 
         for blockno in range(0, self.inodes[fh].size // self.max_obj_size + 1):
-            self.cache.flush(fh, blockno)
+            self.cache.flush_local(fh, blockno)
 
     def forget(self, forget_list):
         log.debug('started with %s', forget_list)
